/* CLParser Version 3.  Copyright (c) Alan Ballard 1991. */

/*****************************************************************************
 *               CLParser Version 3  - clp.h                                 *
 *                                                                           *
 *  Definitions for Clparser routines and types.                             *
 *                                                                           *
 *  This header defines the types etc. comprising the parser routines and    *
 *  structures; it is included in all applications using the parser.         *
 *                                                                           *
 *****************************************************************************/

/* Define some types that are used in following (if they aren't defined
 * already). We'll assume if there is a prior definition it is consistent
 * with our use, since this is fairly conventional....
 * (Note if there is a prior definition via typedef, it is also necessary
 * to define the symbols BOOLDEF or BYTEDEF to suppress this definition.)
 */
#if !defined(BOOLDEF) && !defined(BOOL)
   typedef unsigned short BOOL;
#endif

#if !defined(BYTEDEF) && !defined(BYTE)
   typedef unsigned char BYTE;
#endif

/* For now include md.h to get at definition of UINT.  Want to separate
 * md.h into definitions required by clp.h and those required internally.
 */
#if !defined(MD_INCL)
   #include "md.h"
#endif

/* Types struct _Parser_Control_Block and _Goal_Entry are not really
 * needed in applications using the parser; only the pointers are required
 * as handles.  However, some compilers complain if they are left undefined.
 * The ANSI standard doesn't seem to comment on this.  So we provide 
 * dummy definitions here if the real definitions won't be 
 * included.  The parser and table generator must define CLP_INTERNAL and
 * include internal headers before this one. The tables file must define
 * CLP_TABLES 
 */
#if !defined(CLP_INTERNAL)
   struct _Parser_Control_Block {
      BYTE dummy;
   };
#endif
#if !defined(CLP_INTERNAL) && !defined(CLP_TABLES)
   struct _Goal_Entry {
      BYTE dummy;
   };
#endif

typedef struct _Parser_Control_Block * HPCB; /* Handle for Parser Cont Blk */
typedef const struct _Goal_Entry * GOAL;     /* Goal structure */ 

/* Semantic routines have following form */
typedef BOOL SEMFN(HPCB, void *, UINT);

/* Typedefs for system interface routines ("methods") that are called by the
 * parser.  These are replacable via Parse_Set.  */

/* Routine to read input. Set via Parse_Set(Pcb, Input_Routine, ...)
 * Defaults to fgets; following typedef might have to be made system
 * dependent if any special attributes are required (e.g., Microsoft
 * _cdecl, etc.) */
typedef char * FGETSFN(char *, int, void *);  /* last is usually FILE * */

/* Tracing options. */
typedef enum {Parse_No_Trace, Parse_Semantic_Trace, Parse_Syntax_Trace,
   Parse_All_Trace} PARSE_TRACE;

/* Flag used to determine default tracing state. */
extern PARSE_TRACE Parse_Trace_State;

/* Type defining items that may be set via Parse_Set and/or
 * retrieved via Parse_Get */

/* Note: The order of these must not be changed without
 * recompiling all programs using them. */
typedef enum {
   /* Items from text being parsed */
   Scan_Position, Production_Start_Position, Input_Text,
   Current_Production_Text, Terminal_Text,

   /* Name of current semantic action */
   Semantic_Name,

   /* Error message, error codes */
   Last_Error_Code, Last_Error_Message,

   /* Switches and things */
   Case_Convert, Parse_Trace, Print_Errors,

   /* User storage pointer */
   User_Data,

   /* Version information */
   Tables_Generation_Time,    /* time stamp (time_t) when tables generated. */
   Parser_Version,            /* version of parser */
   Tables_Format_Version,     /* table format */

   /* Routines that may be supplied. */
   Input_Routine,

   /* Miscellaneous interface stuff. */
   Input_IO_Stream,

   /* System Terminals return information */
   Parsed_Integer, Parsed_String, Parsed_Real
}  PARSE_ITEM;


/* Parser routines. */
HPCB Parse_Initialize(
   void * User_Psect);

BOOL Parse_Reset(
   HPCB Pcb);

BOOL Parse_Terminate(
   HPCB Pcb);

BOOL Parse(
   HPCB Pcb,                  /* The private storage */
   GOAL Goal,                 /* Goal production */
   const char * Command,      /* Text to parse (NULL if none) */
   UINT Command_Length);      /* Length of text or 0 if null-terminated. */

BOOL Parse_Argv(
   HPCB Pcb,                  /* The private storage */
   GOAL Goal,                 /* Goal production */
   const char * const Argv[], /* Argv-style string array. */
   int  Argc);                /* Size of Argv. */

INT Parse_Get(
   HPCB Pcb,
   PARSE_ITEM Item_To_Get,
   void * Item_Address,       /* place to return it */
   UINT Space_Length,         /* max to return */
   ...);                      /* optional upper case flag */

BOOL Parse_Set(
   HPCB Pcb,
   PARSE_ITEM Item_To_Set,
   ...);                      /* various things */

