/* CLParser Version 3.  Copyright (c) Alan Ballard 1991. */

/*****************************************************************************
 *               CLParser Version 3  - clptab.h                              *
 *                                                                           *
 *  Internal definitions for Clparser/Clparsegen                             *
 *                                                                           *
 *  This header defines the structures comprising the syntax tables.  The    *
 *  header is required when compiling the parser and the analyzer, and       *
 *  the parsing tables, but **** should not **** be copied by other programs *
 *  since the structure may change                                           *
 *                                                                           *
 *****************************************************************************/

#if !defined(MD_INCL)
   #include "md.h"
#endif

/* Remember this file has been included */
#define CLPTAB_INCL

/* The syntax tables consist of the following pieces:
 *     1) The action list, containing all the syntax items.  The format of 
 *        this is described in the clpint.h header file.  It is emitted by
 *        the analyzer as an array of bytes.
 *     2) The production table, which is indexes into the action list for
 *        the start of each production.
 *     3) The packed character table which is an array of characters.
 *     4) An array of adcons for semantic procedures.
 *     5) An array defining offsets and lengths of production
 *        names in the packed character table.
 *        This is also emitted as an array of bytes.
 *     6) For each semantic procedure, an array giving offsets and
 *        lengths for each semantic name (as in (5)).
 *     7) An array of pointers to the arrays given in (6).
 *     8) A fixed part containing constants and pointers to
 *        other parts.  This is defined by
 *        SYNTAX_TABLE_FIXED_PART.
 *     9) Entry points for each goal.  These consist of
 *        an integer production number, and a pointer to the
 *        fixed part of the tables. See GOAL_ENTRY.
 */

typedef UINT * PRODUCTION_TABLE;    /* array of starting positions */
typedef SEMFN ** SEMFN_TABLE;       /* array of functions */

/* A Symbol_Offset_Array is an array of three byte items (offsets in PCT
 * and length. */
typedef BYTE * SYMBOL_OFFSET_ARRAY;

typedef struct {
   UINT Tables_Version;           /* tables format number */
   const BYTE * Action_List;
   PRODUCTION_TABLE Productions;
   const BYTE * Packed_Characters;
   UINT Number_Nonterminals;
   long Tables_Generation_Time;   /* time stamp when created. Really time_t */
   UINT Number_Goals;
   UINT Number_Productions;
   UINT Number_Terminal_Tables;
   SEMFN_TABLE Adcon_Array;   /* pointer to array of semantic routines */
   SYMBOL_OFFSET_ARRAY Symbol_Offset_Array;  /* array of production names */
   SYMBOL_OFFSET_ARRAY * Semantic_Name_Offset_Array; /* array of arrays */
   } SYNTAX_TABLE_FIXED_PART;

typedef struct _Goal_Entry {
   const SYNTAX_TABLE_FIXED_PART * Fixed_Part;
   UINT Goal_Production_Number;
   } GOAL_ENTRY;

typedef enum {Action_Read_Kind, Action_End_Kind,
   Action_Next_Kind, Action_Fence_Kind,
   Action_Fail_Kind, Action_Succeed_Kind,
   Action_Nonterminal_Link_Kind, Action_Internal_Link_Kind,
   Action_Terminal_Link_Kind,
   Action_Literal_Kind,
   Action_Optional_Literal_Kind, Action_Semantic_Kind,
   Action_Goto_Kind,
   Action_System_Terminal_Kind,
   Action_Quit_Kind,
   Action_Arb_Terminal_Link_Kind,
   Action_One_Char_Term_Link_Kind,
   Action_Table_Terminal_Link_Kind
   } ACTION_KIND;

