/* Copyright (c) 1990-1992 by the University of Illinois Board of Trustees */
/************************************************************************
 * functions for i/o over a pseudo-telnet ctb stream
 * these functions are oriented toward CONVENIENCE, not performance
 ************************************************************************/
#include <stdio.h>
#include <stdarg.h>
#include <StdLib.h>
#include <string.h>
#include <CommResources.h>
#include <Connections.h>

#include <clp.h>

#include "ctb.h"
#include "dlgutil.h"
#include "glob.h"
#include "log.h"
#include "util.h"

/* include files changed by the build process */
#include "script_grammar.h"
#include "script_semantics.h"

#define ZapHandle(aHandle)	do{DisposHandle(aHandle);aHandle=nil;}while(0)
#define PCopy(to,from)	BlockMove(from,to,*(from)+1)
#define userCancelled 	-29999 /* user cancelled operations */

static short	CTBErr;

/************************************************************************
 * private functions
 ************************************************************************/
Boolean CTBExists(void);
short ConfigureCTB(Boolean makeUser);
short CTBNavigateString(Ptr string);
short MyCMOpen(ConnHandle cn,long timeout);
short MyCMClose(ConnHandle cn);
short MyCMWrite(ConnHandle cn, Ptr buffer,long *size);
short MyCMBreak(ConnHandle cn,long howLong);
short MyCMRead(ConnHandle cn,Ptr buffer,long *size,long timeout);
short GetSecondPass(Ptr pass);
pascal void CTBComplete(ConnHandle cn);
long CTBDataCount(void);
typedef struct
{
	short pending;
	short errCode;
	CMBufferSizes asyncCount;
} AsyncStuff, *AsyncStuffPtr;
Boolean SilentCTB=false;
#define CTBTrouble(err) CTBTR(err,0,__LINE__)
short CTBTR(short err,short file,short line);
#define ESC_AUXUSR 1
#define ESC_AUXUSRERR 2
#define ESC_AUXPW 4
#define ESC_USR 8
#define ESC_PW 16
short EscapeCopy(Ptr to,Ptr from,Ptr expect);
short SpinOn(short *rtnCodeAddr,long maxTicks);
void WriteZero(Ptr pointer,long size);

/************************************************************************
 * InitCTB - initialize the CommToolbox
 ************************************************************************/
Boolean InitCTB(Boolean makeUser)
{
	static int CTBInitted=0;
	
	CTBInitted = CTBInitted>4?4:CTBInitted;  /* MIN(CTBInitted,4) */
	CTBSilenceTrans(false);
	switch (CTBInitted)
	{
		case 0:
			if (!CTBExists())
			{
				ErrorMessage("The Communications Toolbox is not installed.");
				return(false);
			}
			CTBInitted++;
		case 1:
			if (CTBErr = InitCTBUtilities())
			{
				ErrorMessage("Couldnt start the Communications Toolbox Utilities.");
				return(false);
			}
			CTBInitted++;
		case 2:
			if (CTBErr = InitCRM())
			{
				ErrorMessage("Couldnt start the Communications Toolbox Resource Manager.");
				return(false);
			}
			CTBInitted++;
		case 3:
			if (CTBErr = InitCM())
			{
				ErrorMessage("Couldnt start the Communications Toolbox Connection Manager.");
				return(false);
			}
			CTBInitted++;
		case 4:
			if (CTBErr = ConfigureCTB(makeUser))
				return(false);
			CTBInitted++;
	}
	return(true);
}

/************************************************************************
 * CTBExists - is the CTB installed?
 ************************************************************************/
Boolean CTBExists(void)
{
#define CTBTrap 0x8b
#define UnimplTrap 0x9f
	return(NGetTrapAddress(UnimplTrap,OSTrap)!=NGetTrapAddress(CTBTrap,OSTrap));
}

/************************************************************************
 * DialThePhone - create a CTB connection
 ************************************************************************/
short DialThePhone(void)
{
	Str255 scratch;
	long ExtraWait = 0;
	
	StatusWindow("Initializing the Communications Toolbox.");
	if (!InitCTB(false)) return (CTBErr?CTBErr:cmGenericError);

	gRcvSpot = -1;
	if ((gRcvBuffer=MyNewHandle(1024))==nil)
	{
		CTBErr=MyMemErr();
		return(CTBErr);
	}
	StatusWindow("Trying to make a connection.");
	if (CTBErr = MyCMOpen(gCnH,-1))
	{
		CMDispose(gCnH);
		gCnH = nil;
		return(CTBErr);
	}
	return(CTBNavigateSTRN(NAVIN));
}
	
/************************************************************************
 * CTBConnectTrans - connect to the remote host.	This version uses the CTB.
 ************************************************************************/
int CTBConnectTrans(Ptr serverName, short port)
{
	Str127 scratch;
	Str127 telnetCommand;
	Handle telnetHandle;
	
	telnetHandle = GetResource('STR ', 1202);
	pstrcpy(telnetCommand, (StringPtr)*telnetHandle);
	p2cstr(telnetCommand);
	sprintf((Ptr)scratch, (Ptr)telnetCommand, serverName, (int) port);
	
	/* Get rid of any extranious data */
	{
    	Ptr buf;
		long bufsize;
		int i;
		Boolean oldSilence = SilentCTB;
		int oldTimeout = CTBTimeout;
		
		gCTBTimeout = 2;	 
		bufsize = 1024;
		if (!(buf=MyNewPtr(bufsize))) return (CTBErr=MyMemErr());
		CTBSilenceTrans(true);
		for (;;) {
		   bufsize = GetPtrSize(buf);
		   if (NetRecvLine(buf, &bufsize)) break;
	  	}
    	DisposPtr(buf);
		CTBSilenceTrans(oldSilence);
		gCTBTimeout = oldTimeout;
  	}
	CTBSendTrans(1,(Ptr)scratch,strlen((Ptr)scratch));
		
	return(CTBErr);
}

/************************************************************************
 * CTBSendTrans - send some text to the remote host.	This version uses the CTB.
 ************************************************************************/
int CTBSendTrans(short count, Ptr text,long size, ...)
{
	long bSize;
	if (size==0) return(noErr); 	/* allow vacuous sends */
	GiveTime();
	
	bSize = size;
	if (CTBErr=MyCMWrite(gCnH,text,&bSize))
		return(CTBErr);
		
  /* If we need to handle echo then do a read for each carriage return. */
  if (gScriptGlobals.TrapEcho) {
     Ptr buf;
	 long bufsize;
	 int i;
	 Boolean oldSilence = SilentCTB;
	 int oldTimeout = gCTBTimeout;
	 
	 gCTBTimeout = 2;
	 bufsize = 1024;
	 if (!(buf=MyNewPtr(bufsize))) return (CTBErr=MyMemErr());
	 CTBSilenceTrans(true);
	 for (i=0; i<size && !CTBErr; i++)
	   if (strncmp(text+i, gCRLF, strlen(gCRLF)) == 0) {
		   bufsize = GetPtrSize(buf);
		   if (NetRecvLine(buf, &bufsize)) break;
		   i += strlen(gCRLF)-1;
	   }
	   
	 gCTBTimeout = oldTimeout;
	 
     DisposPtr(buf);
	 CTBSilenceTrans(oldSilence);
  }
  
	if (--count>0)
	{
		Ptr buffer;
		va_list extra_buffers;
		va_start(extra_buffers,size);
		while (count--)
		{
			GiveTime();
			buffer = va_arg(extra_buffers,Ptr);
			bSize = va_arg(extra_buffers,int);
			if (CTBErr=CTBSendTrans(1,buffer,bSize)) return CTBErr;
		}
		va_end(extra_buffers);
	}
	return(noErr);
}

/************************************************************************
 * GenSendWDS - send a lot of text to the remote host.
 ************************************************************************
int GenSendWDS(wdsEntry *theWDS)
{
	short err=0;

	GiveTime();
	for (;theWDS->length;theWDS++)
		if (err=SendTrans(1,theWDS->ptr,theWDS->length)) break;
	
	return(err);
}
*/

/************************************************************************
 * CTBRecvTrans - get some text from the remote host.  This version uses the CTB.
 ************************************************************************/
int CTBRecvTrans(Ptr line,long *size)
{
	long gotThis;
	Str31 tmStr;
	short factor = 60;
	Ptr spot = line;
	long want = *size;
	
	NumToString(gCTBTimeout,tmStr);
	MyCMIdle();
	do
	{
		CTBErr = SpinOn(&gCTBHasChars,gCTBTimeout*factor);
		if (CTBErr)
			CTBErr = cmTimeOut;
		else
		{
			factor = 10;
			gotThis = CTBDataCount();
			gotThis = (gotThis>want)?want:gotThis; /* MIN(gotThis,want) */
			CTBErr=MyCMRead(gCnH,line,&gotThis,0);
			if (!CTBErr) {spot += gotThis; want -= gotThis;}
		}
	}
	while (want && spot==line && !SilentCTB && !gCancel
				 /* && AlertStr(TIMEOUT_ALRT,Caution,tmStr)==1 */);

	*size = spot - line;
	if (!CTBErr && *size==0) CTBErr = cmTimeOut;
	return(CTBErr);
}

/************************************************************************
 *
 ************************************************************************/
void HangUpThePhone(void)
{
	if (gCnH)
	{
		Str127 msg;
		(void) CTBNavigateSTRN(NAVOUT);
		StatusWindow("Closing the connection.");
		/* if (!PrefIsSet(PREF_NO_HANGUP)) */
			(void) MyCMClose(gCnH);
	}
	if (gCnH)
	{
		CMDispose(gCnH);
		gCnH = nil;
	}
	if (gRcvBuffer) ZapHandle(gRcvBuffer);
}

/************************************************************************
 * CTBDisTrans - disconnect from the remote host.  This version uses the CTB.
 ************************************************************************/
int CTBDisTrans(void)
{
	return(noErr);
}

/************************************************************************
 * CTBDestroyTrans - destory the connection.
 ************************************************************************/
int CTBDestroyTrans(void)
{
	return(noErr);
}

/************************************************************************
 * CTBTransError - report our most recent error
 ************************************************************************/
int CTBTransError(void)
{
	return(CTBErr);
}

/************************************************************************
 * CTBSilenceTrans - turn off error reports from tcp routines
 ************************************************************************/
void CTBSilenceTrans(Boolean silence)
{
	SilentCTB = silence;
}

/************************************************************************
 * ConfigureCTB - configure the communications toolbox
 ************************************************************************/
short ConfigureCTB(Boolean makeUser)
{
	Str63 toolName;
	short toolNameID;
	Str255 scriptName;
	short procID;
	CMBufferSizes bs;
	Point pt;
	short result;
	Handle cfigH;
	short cfigHID;
	scriptInfoPtr theScript;
	short scriptCount;
	Boolean needConfig = true;
	ConnHandle	oldCnH;
	
	CTBErr = 0;
	
	/*
	 * Get script name 
	 */
	 
	HLock((Handle)gOurScripts);
	scriptName[0]=0;
	for (theScript=*gOurScripts,scriptCount=1; scriptCount<gScriptNumber;
		scriptCount++,theScript = (scriptInfoPtr)(theScript->ourLength+(long)theScript));
	if (scriptCount>0) PCopy(scriptName,theScript->fullName);
	HUnlock((Handle)gOurScripts);
	
	/*
	 * find the tool to use
	 * If using a script, get the tool associated with the script
	 */
	if (gScriptNumber < 1) {
		Handle strH = (Handle)GetString(1001);
		if (!strH) return(1);
		PCopy(toolName,*strH);
		ReleaseResource(strH);
	}
	else {
		Handle toolNameH;
		ResType toolNameRes;
		Str255 toolNameName;
		toolName[0]=0;
		toolNameID=0;
		if (toolNameH=GetNamedResource('CTnm',scriptName)) {
			PCopy(toolName,*toolNameH);
			GetResInfo(toolNameH,&toolNameID,&toolNameRes,toolNameName);
			ReleaseResource(toolNameH);
		}
	}
	
	if ((procID=CMGetProcID(toolName))<0) {
		CRMGetIndToolName(classCM,1,toolName);
		if (!*toolName) {
			ErrorMessage("No connection tools are available.");
			return(1);
		}
		else if ((procID=CMGetProcID(toolName))<0) {
			ErrorMessage("Couldnt get connection tool.");
			return(1);
		}
	}
	
	/*
	 * create a connection record
	 */
	WriteZero((Ptr)bs,sizeof(bs));
	oldCnH = gCnH;
	if (!(gCnH = CMNew(procID,cmNoMenus,bs,0,0)))
	{
		ErrorMessage("The operation failed; there was not enough memory.");
		return(1);
	}
	
	/*
	 * try to load the old configuration
	 */
	if (gScriptNumber < 1) cfigH = GetNamedResource('CTfg',toolName);
	else {
		cfigH = GetNamedResource('CTfg',scriptName);
		cfigHID = 0;
		if (cfigH) {
			ResType cfigHType;
			Str255 cfigHName;
			GetResInfo(cfigH,&cfigHID,&cfigHType,cfigHName);
		}
	}
	if (cfigH)
	{
		if (GetPtrSize((*gCnH)->config)==GetHandleSize(cfigH))
		{
			BlockMove(*cfigH,(*gCnH)->config,GetPtrSize((*gCnH)->config));
			BlockMove(*cfigH,(*gCnH)->oldConfig,GetPtrSize((*gCnH)->config));
			needConfig = CMValidate(gCnH);
		}
		ReleaseResource(cfigH);
	}
	
	/*
	 * let the user have a look at things
	 */
	if (makeUser || needConfig)
	{
		pt.h = 10; pt.v = GetMBarHeight()+10;
		switch (result = CMChoose(&gCnH,pt,nil))
		{
			case chooseDisaster:
			case chooseFailed:
			case chooseAborted:
				CTBErr = result;
				break;
			case chooseOKMajor:
			case chooseOKMinor:
				/* save the name string */
				CMGetToolName((*gCnH)->procID,toolName);
				BlockMove((*gCnH)->config,(*gCnH)->oldConfig,GetPtrSize((*gCnH)->config));
				if (gScriptNumber < 1) CTBErr = SettingsPtr('STR ',nil,1001,(Ptr)toolName,*toolName+1);
				else {
					if (!toolNameID) while (128>(toolNameID=UniqueID('CTnm')));
					CTBErr = SettingsPtr('CTnm',(Ptr)scriptName,toolNameID,(Ptr)toolName,*toolName+1);
				}
				if (!CTBErr) {
					if (gScriptNumber < 1) 
						CTBErr = SettingsPtr('CTfg',(Ptr)toolName,1001,(*gCnH)->config,GetPtrSize((*gCnH)->config));
					else {
						if (!cfigHID) while (128>(cfigHID=UniqueID('CTfg')));
						CTBErr = SettingsPtr('CTfg',(Ptr)scriptName,cfigHID,(*gCnH)->config,GetPtrSize((*gCnH)->config));
					}
				}
				if (CTBErr) ErrorMessage("Couldn't save your settings.");
				break;
		}
	}
	
	/*
	 * just configuring?
	 */
	if (gCnH)
		if (makeUser) {CMDispose(gCnH); gCnH = oldCnH;}
	
	return(CTBErr);
}

/************************************************************************
 * CTBNavigateSTRN - navigate a serial connection by means of an STRN resource
 ************************************************************************/
short CTBNavigateSTRN(short id)
{
	short count,n;
	Str63 name;
	Handle strnH;
	Str255 string;
	ResType theType;
	Boolean oldSilence = SilentCTB;
	short err=noErr;
	scriptInfoPtr thisScript;
	
	/* First check to see if the user has selected a script, and if so, if that
		script is available. */
	
	if (*gScriptName) {
		HLock((Handle)gOurScripts);
		
		/* The user has selected a script, now we will see if it exists */
		for (thisScript=*gOurScripts; thisScript->ourLength; 
			thisScript = (scriptInfoPtr)(thisScript->ourLength+(long)thisScript) )
				if (EqualString((StringPtr)(thisScript->fullName), gScriptName, false, true))
					break;
					
		/* If the entry has a length, then we have a script */
		if (thisScript->ourLength) {
			short theID;
			ResType theTYPE;
			short oldRefN = CurResFile();
			
			/* Look for the NAVIN, NAVMID or NAVOUT etc. script */
			UseResFile(thisScript->scriptRefN);
			strnH = GetNamedResource('TEXT', gScriptName);
			if (id!=NAVIN) {
				GetResInfo(strnH, &theID, &theTYPE, gScriptName);
				theID += id==NAVMID    ? 1 :
						 id==NAVOUT    ? 2 :
						 id==NAVTELIN  ? 3 :
					/*	 id==NAVTELOUT*/ 4;
				ReleaseResource(strnH);
				strnH = Get1Resource('TEXT', theID);
			}
			
			UseResFile(oldRefN);
			
			/* If there is one, use it. */
			if (strnH) {
				HPCB Pcb;
				Handle lineEndHandle;
	
				Pcb = Parse_Initialize(&gScriptGlobals);
				gScriptGlobals.stack = (PScript_Stack_Frame)NewPtrClear(sizeof(Script_Stack_Frame));
				gScriptGlobals.stack->doCommands = true;
				gScriptGlobals.TraceRefN = false;
				gScriptGlobals.ParseRC = noErr;
				gScriptGlobals.ourRefN = thisScript->scriptRefN;
		
				HLock(strnH);
				CTBSilenceTrans(true);
				gScriptGlobals.Script_Length = SizeResource(strnH);
				if (!Parse(Pcb, script, *strnH, gScriptGlobals.Script_Length))
					gScriptGlobals.ParseRC = ParseQuit;
				CTBSilenceTrans(oldSilence);
				gCTBTimeout = 45;
				if (gScriptGlobals.TraceRefN) fclose(stdout);
				FreeStack(gScriptGlobals.stack);

				HUnlock(strnH);
				ReleaseResource(strnH);
				Parse_Terminate(Pcb);
				lineEndHandle = GetResource('STR ', 1201);
				pstrcpy((StringPtr)gCRLF, (StringPtr)*lineEndHandle);
				p2cstr((StringPtr)gCRLF);
				ReleaseResource(lineEndHandle);
			} 
			
			HUnlock((Handle)gOurScripts);
			return CTBErr = gScriptGlobals.ParseRC;
		}
		HUnlock((Handle)gOurScripts);
	}
	
	/* Didn't find a script, so do the standard Eudora navigation thing. */
	strnH = GetResource('STR ', id);
	pstrcpy(name, (StringPtr)*strnH);
	CMGetToolName((*gCnH)->procID,string);
	p2cstr(string);
	p2cstr(name);
	strcat((Ptr)string, " ");
	strcat((Ptr)string, (Ptr)name);
	c2pstr((Ptr)string);
	c2pstr((Ptr)name);
	strnH = GetNamedResource('STR#',string);
	if (!strnH)
		strnH = GetNamedResource('STR#',name);
	if (!strnH)
		strnH = GetResource('STR#',id==NAVIN ? CTB_NAV_IN_STRN :
			(id==NAVMID         ? CTB_NAV_MID_STRN : 
			(id==NAVOUT         ? CTB_NAV_OUT_STRN :
			(id==NAVTELIN       ? CTB_NAV_TEL_IN_STRN :
			(/*id==NAVTELOUT ?*/  CTB_NAV_TEL_OUT_STRN
			)))));
	
	if (strnH)
	{
		gCTBTimeout = 2;
		count = (*strnH)[0]*256 + (*strnH)[1];
		GetResInfo(strnH,&id,&theType,name);
		ReleaseResource(strnH);
		
		CTBSilenceTrans(true);
		for (n=1; n<=count; n++)
		{
			GetIndString(string,id,n);
			if (err=CTBNavigateString((Ptr)string)) break;
		}
		CTBSilenceTrans(false);
		gCTBTimeout = 45;
		if (err!=cmTimeOut) CTBTrouble(err);
	}
	return(err==cmTimeOut ? noErr : err);
}

/************************************************************************
 * CTBNavigateString - navigate a serial connection by a string
 ************************************************************************/
short CTBNavigateString(Ptr string)
{
	Str255 data, scratch, expect;
	long bSize;
	long tics;
	Ptr exSpot,dSpot;
	Boolean found=false,brk=false,wait=false;
	short escFlags;
	
	escFlags = EscapeCopy((Ptr)data,(Ptr)string,(Ptr)expect);
	for (exSpot=(Ptr)expect+*expect;exSpot>(Ptr)expect;exSpot--) *expect &= 0x7f;	/* strip parity */
	
	if (gCancel) return(userCancelled);
	if (escFlags & ESC_AUXUSRERR)
	{
		ErrorMessage("Your navigation script requires a dialup username.  Use the Configuration... dialog to set one.");
		return(userCancelled);
	}
	if (*data && data[1]==bulletChar)
	{
		StatusWindow("secret things");
		BlockMove(data+2,data+1,*data-1);
		--*data;
	}
	else if (*data && ((brk=data[1]==betaChar) || (wait=data[1]==deltaChar)))
	{
		BlockMove(data+2,data+1,*data);
		--*data;
		tics = atoi((Ptr)data+1);
		if (wait) {
			 tics *= 60;
			StatusWindow("Waiting...");
		}
		else {
			sprintf((Ptr)scratch, "BREAK [%d ticks]", tics);
			StatusWindow((Ptr)scratch);
		}
	}
	else
	{
		StatusWindow((Ptr)data+1);
	}

	if (brk)
	{
		if (CTBErr=MyCMBreak(gCnH,tics)) return(CTBErr);
	}
	else if (wait)
	  for (tics+=TickCount();TickCount()<tics;) GiveTime();
	else {
		LogNavigateSent((Ptr)data+1, *data);
		if (CTBErr=CTBSendTrans(1,(Ptr)data+1,*data)) return(CTBErr);
	}
	if (*expect)
	{
		tics = TickCount()+45*60;
		exSpot = (Ptr)expect+1;
	}
	do
	{
		bSize = sizeof(data)-1;
		CTBErr = NetRecvLine((Ptr)data+1,&bSize);
		if (bSize)
		{
			data[0] = bSize;
			StatusWindow((Ptr)data+1);
			if (gPrefs.logActionsToFile) {
				LogNavigateResp((Ptr)data+1, bSize);
			}
			if (*expect)
			{
				for(dSpot=(Ptr)data+1;dSpot<=(Ptr)data+*data;dSpot++)
					if ((*dSpot&0x7f)==*exSpot)
					{
						if (++exSpot>(Ptr)expect+*expect)
						{
							found = true;
							break;
						}
					}
					else
						exSpot = (Ptr)expect+1;
			}
		}
		if (*expect && TickCount()>tics)
		{
				CTBErr = userCancelled;
		}
	}
	while (!gCancel && (!CTBErr || CTBErr==cmTimeOut&&*expect) && !found);
	if (CTBErr==cmTimeOut) CTBErr = 0;
	if (*expect && !found)
	{
		CTBErr = userCancelled;
	}
	return(CTBErr);
}

/************************************************************************
 * MyCMClose - call CMClose
 ************************************************************************/
short MyCMClose(ConnHandle cn)
{
	AsyncStuff stuff;
	short err;						/* real error */

	(*cn)->refCon = (long) &stuff;				/* store address */
	stuff.pending = 1;
	if (true) /* syncronous */
		err=CMClose(cn,false,nil,-1,true);
	else
	{
		(void) CMIOKill(cn,cmDataIn);
		(void) CMIOKill(cn,cmDataOut);
		if (!(err=CMClose(cn,true,CTBComplete,-1,true)))
			err = SpinOn(&stuff.pending,0) || stuff.errCode;
	}
	CTBTrouble(err);
	return (err); 				/* error? */
}

/************************************************************************
 * MyCMOpen - call CMOpen
 ************************************************************************/
short MyCMOpen(ConnHandle cn,long timeout)
{
	AsyncStuff stuff; 										
	short err;						/* real error */
	
	(*cn)->refCon = (long) &stuff;				/* store address */
	stuff.pending = 1;
	if (true) /* sync */
		err=CMOpen(cn,false,CTBComplete,timeout);
	else
	{
		if (!(err=CMOpen(cn,true,CTBComplete,timeout)))
		{
			err = SpinOn(&stuff.pending,0) || stuff.errCode;
			if (err==userCancelled)
				(void) CMAbort(gCnH);
		}
	}

	CTBTrouble(err);
	return (err);
}

/************************************************************************
 * MyCMRead - call CMRead
 ************************************************************************/
short MyCMRead(ConnHandle cn,Ptr buffer,long *size,long timeout)
{
	AsyncStuff stuff;
	short err;						/* real error */
	CMFlags junk;
	
	(*cn)->refCon = (long)&stuff;
	stuff.pending = 1;
	if (!(err=CMRead(cn,buffer,size,cmData,true,CTBComplete,timeout,&junk)))
	{
		err = SpinOn(&stuff.pending,0) || stuff.errCode;
		if (err==userCancelled)
			CMIOKill(cn,cmDataIn);
	}

	if (!err || err==cmTimeOut)
	{
		*size = stuff.asyncCount[cmDataIn];
		if (*size) err = noErr;
		else err = cmTimeOut;
	}
	else
		*size = 0;
	if (err!=cmTimeOut) CTBTrouble(err);
	return(err);
}

/************************************************************************
 * MyCMWrite - call CMWrite
 ************************************************************************/
short MyCMWrite(ConnHandle cn,Ptr buffer,long *size)
{
	AsyncStuff stuff;
	short err;						/* real error */
	(*cn)->refCon = (long)&stuff;
	stuff.pending = 1;
	/* if (!DontTranslate && TransOut)
		{ptr cp; for(cp=buffer;cp<buffer+*size;cp++) *cp = TransOut[*cp];} */

		if (gPrefs.logActionsToFile) {
			LogNavigateSent(buffer, *size);
		}

	if (!(err=CMWrite(cn,buffer,size,cmData,true,CTBComplete,-1,0)))
	{
		err = SpinOn(&stuff.pending,0) || stuff.errCode;
		if (err==userCancelled)
			CMIOKill(cn,cmDataOut);
	}

	if (!err)
		*size = stuff.asyncCount[cmDataOut];
	else
		*size = 0;

	CTBTrouble(err);
	return(err);
}

/************************************************************************
 * MyCMBreak - call CMBreak
 ************************************************************************/
short MyCMBreak(ConnHandle cn,long howLong)
{
	AsyncStuff stuff;
	short err;						/* real error */
	(*cn)->refCon = (long)&stuff;
	stuff.pending = 1;
	CMBreak(cn,howLong,true,CTBComplete);
	
	err = SpinOn(&stuff.pending,0) || stuff.errCode;
	if (err==userCancelled)
		CMIOKill(cn,cmDataOut);

	CTBTrouble(err);
	return(err);
}

/************************************************************************
 * CTBTR - report a CTB error
 ************************************************************************/
short CTBTR(short err,short file,short line)
{
	Handle strnH;
	
	if (!SilentCTB &&!gCancel && err && err!=userCancelled)
	{
		strnH = GetResource('STR#',3600);
		if (strnH)
		{
			char *errString;
			Str127 message;
			strcpy((Ptr)message, "Dont blame me; blame the CTB. ");
			if (err+2 > 0 && err+2 <= (*strnH)[0]*256 + (*strnH)[1]) {
				errString = *strnH+2;
				for (err +=2; err > 0;err--)
					errString += (*errString)+1;
				strncat((Ptr)message, errString+1, *errString);
			}
			else {
				Str127 strNum;
				NumToString(err,strNum);
				p2cstr(strNum);
				strcat((Ptr)message, (Ptr)strNum);
			}
			ReleaseResource(strnH);
			ErrorMessage((Ptr)message);
			CTBSilenceTrans(true);
		}
	}
	return(err);
}

/************************************************************************
 * CTBWhoAmI - return the mac's ctb name
 ************************************************************************/
Ptr CTBWhoAmI(Ptr who)
{
	strcpy(who, "DialupNewsWatcher");
	return(who);
}

/************************************************************************
 * CTBBreak - send a break
 ************************************************************************/
int CTBBreak(long duration)
{
  return MyCMBreak(gCnH, duration);
}

/************************************************************************
 * MyCMIdle - do the things I should do at CMIdle time
 ************************************************************************/
void MyCMIdle(void)
{
	CMIdle(gCnH);
	gCTBHasChars = CTBDataCount() ? noErr : inProgress;
}

/************************************************************************
 * CTBDataCount - how much data do we have?
 ************************************************************************/
long CTBDataCount(void)
{
	CMStatFlags flags;
	CMBufferSizes sizes;
	
	CMStatus(gCnH,sizes,&flags);
	return(flags&cmStatusDataAvail ? sizes[cmDataIn] : 0);
}

/************************************************************************
 * GetSecondPass - get the user's secondary password
 ***********************************************************************
short GetSecondPass(Ptr pass)
{
	Str31 save;
	short err;
	Str31 dialup;
	
	if (!*SecondPass)
	{
		PCopy(save,Password);
		*Password = 0;
		GetRString(dialup,DIALUP);
		if (!(err=GetPassword(dialup,pass,32,ENTER)))
		{
			PCopy(SecondPass,pass);
			if (PrefIsSet(PREF_SAVE_PASSWORD) && *SecondPass)
			{
				ChangeStrn(PREF_STRN,PREF_AUXPW,Password);
				UpdateResFile(SettingsRefN);
			}
		}
		PCopy(Password,save);
		return(err);
	}
	else
	{
		PCopy(pass,SecondPass);
		return(noErr);
	}
}
*/

/************************************************************************
 * EscapeCopy - copy a string, expanding escapes
 ************************************************************************/
short EscapeCopy(Ptr to,Ptr from,Ptr expect)
{
	Ptr end = from+*from;
	Ptr spot = to+1;
	Boolean esc;
	Str127 scratch1, scratch2;
	short returnVal=0;
	
	*expect = 0;
	from++;
	for (esc=false; from<=end; from++)
	{
		if (esc)
		{
			esc = false;
			switch (*from)
			{
				case 'U':
					BlockMove(gDialupUsername, spot, strlen(gDialupUsername));
					spot += strlen(gDialupUsername);
					break;
				case 'P':
					if (!(*gDialupPassword)) {
						ResType ourResType;
						Str255 ourResName;
						short ourResNumber;
						char password[32];
						char userName[128];
						DialogPtr myDialog;
						Handle ourResource = GetNamedResource('DLOG',"\pPassword");
						strcpy(userName,gDialupUsername);
						c2pstr(userName);
						ParamText((StringPtr)userName,"\pEnter","\p","\p");
						GetResInfo(ourResource, &ourResNumber, &ourResType, ourResName);
						myDialog = MyGetNewDialog(ourResNumber);
						password[0]=0;
						DlgSetCString(myDialog, 4, "");
						SetItemPassword(myDialog, 4, password);
						SetItemMaxLength(myDialog, 4, 31);
						SetItemFlash(myDialog, 5, true);
						for (;;) {
							short itemHit;
							MyModalDialog((ModalFilterProcPtr)DialogFilter, &itemHit, true, true);
							if (itemHit == ok) strcpy(gDialupPassword, password);
							if (itemHit == ok || itemHit == cancel) break;
						}
						MyDisposDialog(myDialog);
						if (!(*gDialupPassword)) {
							gCancel = true;
							break;
						}
					}
					BlockMove(gDialupPassword,spot,strlen(gDialupPassword));
					spot += strlen(gDialupPassword);
					returnVal |= ESC_PW;
					break;
				case 'u': 
					BlockMove(gPrefs.name, spot, strlen(gPrefs.name));
					spot += strlen(gPrefs.name);
					returnVal |= ESC_USR;
					break;
				case 'h': 
					BlockMove(gPrefs.host, spot, strlen(gPrefs.host));
					spot += strlen(gPrefs.host);
					break;
				case 's':
					BlockMove(gPrefs.mailServerName+1,spot,*gPrefs.mailServerName);
					spot += *gPrefs.mailServerName;
					break;
				case 'e':						/* switch to 'expect' string */
					*to = spot-to-1;
					to[*to+1] = 0;
					to = expect;
					spot=to+1;
					break;
				case 'p':
					if (!(*gPrefs.remotePassword)) {
						ResType ourResType;
						Str255 ourResName;
						short ourResNumber;
						char password[32];
						char userName[128];
						DialogPtr myDialog;
						Handle ourResource = GetNamedResource('DLOG',"\pPassword");
						strcpy(userName,gPrefs.name);
						c2pstr(userName);
						ParamText((StringPtr)userName,"\pEnter","\p","\p");
						GetResInfo(ourResource, &ourResNumber, &ourResType, ourResName);
						myDialog = MyGetNewDialog(ourResNumber);
						password[0]=0;
						DlgSetCString(myDialog, 4, "");
						SetItemPassword(myDialog, 4, password);
						SetItemMaxLength(myDialog, 4, 31);
						SetItemFlash(myDialog, 5, true);
						for (;;) {
							short itemHit;
							MyModalDialog((ModalFilterProcPtr)DialogFilter, &itemHit, true, true);
							if (itemHit == ok) strcpy(gPrefs.remotePassword, password);
							if (itemHit == ok || itemHit == cancel) break;
						}
						MyDisposDialog(myDialog);
						if (!(*gPrefs.remotePassword)) {
							gCancel = true;
							break;
						}
					}
					BlockMove(gPrefs.remotePassword,spot,strlen(gPrefs.remotePassword));
					spot += strlen(gPrefs.remotePassword);
					returnVal |= ESC_PW;
					break;
				case 'n': *spot++ = '\r'; break;
				case 'r': *spot++ = '\n'; break;
				case 'b': *spot++ = bulletChar; break;
				case 'B': *spot++ = betaChar; break;
				case 'D': *spot++ = deltaChar; break;
				default: *spot++ = *from; break;
			}
		}
		else
		{
			esc = *from=='\\';
			if (!esc) *spot++ = *from;
		}
	}
	*to = spot-to-1;
	to[*to+1] = 0;	/* null terminate, just for fun */
	return(returnVal);
}


/************************************************************************
 * CTBComplete - completion routine
 ************************************************************************/
pascal void CTBComplete(ConnHandle cn)
{
	AsyncStuffPtr stuffP = (AsyncStuffPtr)(*cn)->refCon;
	long *fp, *tp, *ep;
	
	stuffP->errCode = (*cn)->errCode;
	fp=(*cn)->asyncCount;
	tp=stuffP->asyncCount;
	ep=tp+sizeof(CMBufferSizes)/sizeof(long);
	while (tp<ep) *tp++ = *fp++;
	stuffP->pending = 0;
}

/************************************************************************
 * RecvLine - get some text from the remote host.
 ************************************************************************/
int NetRecvLine(Ptr line,long *size)
{
	long count;
	Ptr anchor, end;
	short err;
	static int RcvBufferSize;
	long bSize = *size;
	Byte c;
	
	*size = 0;
	if (!RcvBufferSize) RcvBufferSize = GetHandleSize(gRcvBuffer);

	anchor = line;
	end = line+bSize-1;
	while (anchor < end) 										/* need more chars */
		if (gRcvSpot>=0) 											/* there are some buffered chars */
		{
			Ptr rPtr = *gRcvBuffer+gRcvSpot;
			
			for(c=*rPtr++;anchor<end;c=*rPtr++)
				if (c /* && c!='\n' */ && !(c==CR&&anchor[-1]==CR))
				{
			 		*anchor++ = c;
			 		if (c==LF) break;
					/* if (c=='\r')
						{
							anchor[-1] = '\n';
							break;
						}
					*/
				}
			
			gRcvSpot = rPtr - *gRcvBuffer;
			if (gRcvSpot>gRcvSize)					/* newline was sentinel */
				anchor--;
			if (gRcvSpot>=gRcvSize)					/* we have emptied the buffer */
				gRcvSpot = -1;
			if (anchor[-1]==/*'\n'*/LF || anchor>=end)	/* found a valid newline, or filled buffer */
			{
				*size = anchor-line;
				*anchor = 0;										/* terminate */
		if (gPrefs.logActionsToFile) {
			LogNavigateResp((Ptr)line, *size);
		}
				return(noErr);
			}
		}
		else
		{
			count = RcvBufferSize-1;
			if (err=CTBRecvTrans(LDRef(gRcvBuffer),&count)) 	/* get some chars */
				 count=0;
			UL(gRcvBuffer);
			if (count)
			{
				(*gRcvBuffer)[count] = /*'\r'*/LF; 						/* sentinel */
				gRcvSize = count;
				gRcvSpot = 0;
			}
			if (err)
			{
				*size = anchor-line;
				line[*size] = 0;
				return(err);
			}
		}
	*size = anchor - line;
	if (*size) line[*size] = 0; 					/* null terminate */
	
		if (gPrefs.logActionsToFile) {
			LogNavigateResp((Ptr)line, *size);
		}
	
	return(*size ? noErr : err);
}

/************************************************************************
 * SpinOn - spin until a return code is not inProgress or cacheFault
 ************************************************************************/
short SpinOn(short *rtnCodeAddr,long maxTicks)
{
	long ticks=TickCount();
	long startTicks=ticks+120;
	long now;
	EventRecord event;
	Boolean oldCommandPeriod = gCancel;
	
	gCancel = false;
	do
	{
		now = TickCount();
		if (now>startTicks && now-ticks>10) {GiveTime();ticks=now;}
		if (gCancel) return(userCancelled);
		MyCMIdle();
		if (maxTicks && startTicks+maxTicks < now+120) break;
	}
	while (*rtnCodeAddr == inProgress || *rtnCodeAddr == cacheFault);
	gCancel = oldCommandPeriod;
	return(*rtnCodeAddr);
}

/**********************************************************************
 * write zeroes over an area of memory
 **********************************************************************/
void WriteZero(Ptr pointer,long size)
{
	while (size--) *pointer++ = 0;
}

/************************************************************************
 * SettingsPtr - put some data into the Settings file
 ************************************************************************/
short SettingsPtr(ResType type,Ptr name,short id,Ptr dataPtr, long dataSize)
{
	Handle rH = MyNewHandle(dataSize);
	short oldRefN = CurResFile();
	short err;
	Handle oldDataHandle;
	
	if (!rH) return(MyMemErr());
	BlockMove(dataPtr,*rH,dataSize);

	UseResFile(gSettingsRefN);
	oldDataHandle = Get1Resource(type,id);
	if (oldDataHandle!=nil)
	{
		RmveResource(oldDataHandle);
		DisposHandle(oldDataHandle);
	}
	
	AddResource(rH,type,id,(StringPtr)(name?name:""));
	err = ResError();
	UseResFile(oldRefN);
	if (err)
		DisposHandle(rH);
	return(err);
}
