/*----------------------------------------------------------------------------

	init.c

	This module handles initialization of the program.
	
	Portions copyright  1990, Apple Computer.
	Portions copyright  1993, Northwestern University.

----------------------------------------------------------------------------*/

#include <GestaltEqu.h>
#include <Traps.h>
#include <Folders.h>
#include <stdio.h>
#include <string.h>

#include "glob.h"
#include "init.h"
#include "menus.h"
#include "util.h"
#include "dlgutil.h"
#include "aevt.h"
#include "prefs.h"
#include "nntp.h"
#include "full.h"
#include "newsrc.h"
#include "killfile.h"
#include "log.h"

/*----------------------------------------------------------------------------
	FindScripts 
	
	Creates a structure of the available scripts.	 
----------------------------------------------------------------------------*/

static void FindScripts(void)
{
	int ResCount = CountResources((ResType)'TEXT');
	int ResNumber;
		
	/* initialize ourScripts with a zero length dummy script */
		
	gOurScripts = (scriptInfoHandle)NewHandleClear(sizeof(short));
	
	/* Initialize some other script related things */
	gScriptGlobals.TrapEcho = false;
	gScriptGlobals.Echo = true;
		
	/* check all TEXT resources to see if they are scripts */ 
		
	for (ResNumber=1; ResNumber<=ResCount; ResNumber++) {
		short theID;
		ResType theType;
		Str255 theName;
		short entryLength;
		Handle theRes=GetIndResource((ResType)'TEXT', ResNumber);
		GetResInfo(theRes, &theID, &theType, theName);
		theName[theName[0]+1] = 0; /* make sure the pascal string is also a c one */
		
		/* if the resource has a name and it contains the string "Script" then bingo */
			
		if (theName[0] && strstr((char *)&theName[1], "Script")) {
			scriptInfoPtr newScript=(scriptInfoPtr)(GetHandleSize((Handle)gOurScripts)-sizeof(short));
			entryLength = sizeof(scriptInfo) + theName[0];
			entryLength += entryLength & 1;  /* Make sure the length is even */
			SetHandleSize((Handle)gOurScripts, GetHandleSize((Handle)gOurScripts) + entryLength);
			if (noErr == MemError()) {
				HLock((Handle)gOurScripts);
				newScript = (scriptInfoPtr)((long)*gOurScripts + (long)newScript);
				newScript->ourLength = entryLength;
				newScript->scriptRefN = HomeResFile(theRes);
				BlockMove(theName, newScript->fullName, theName[0]+2);
				newScript->printNameOffset = ((long)strstr((char *)&theName[1],"Script")) - ((long)&theName[0]) + 6;
				newScript->printNameOffset += strspn((char *)theName+newScript->printNameOffset, " ");
				/* finally, reset the dummy script info at the end to zero length */
				newScript=(scriptInfoPtr)((long)*gOurScripts + GetHandleSize((Handle)gOurScripts) - sizeof(short));
				newScript->ourLength = 0;
				HUnlock((Handle)gOurScripts);
			}
		}
	}
}

/************************************************************************
 * OpenPlugIns - open extra Eudora type resource files
 ************************************************************************/
static void OpenPlugIns(void)
{
	short vRef;
	long dirId;
	HFileInfo hfi;
	Str31 name;
	if (!FindFolder(kOnSystemDisk,kPreferencesFolderType,false,&vRef,&dirId))
	{
		hfi.ioNamePtr = name;
		for (hfi.ioFDirIndex = 1; ;hfi.ioFDirIndex++) {
			hfi.ioDirID = dirId;
			hfi.ioVRefNum = vRef;
			if (PBGetCatInfo((CInfoPBPtr)&hfi,false)) break;
			if (hfi.ioFlFndrInfo.fdType=='rsrc' && hfi.ioFlFndrInfo.fdCreator=='CSOm')
				HOpenResFile(vRef, dirId, name, fsRdPerm);
		}
	}
}

/*----------------------------------------------------------------------------
	SetUpMenus 
	
	Draws the pull-down menus onto the screen.	 
----------------------------------------------------------------------------*/

static void SetUpMenus (void)
{
	Handle mbar;
	
	mbar = GetNewMBar(kMBarID);
	SetMenuBar(mbar);
	AddResMenu(GetMHandle(kAppleMenu), 'DRVR');
	
	#ifndef FILTERS
	{
		short item;
		MenuHandle theMenu;
		Str255 theCommand;
		
		theMenu = GetMHandle(kSearchMenu);
		for (item = kGroupKillItem; item <= kKillSubjectItem; item++) {
			GetItem(theMenu, item, theCommand);
			BlockMove(theCommand+1, theCommand+2, *theCommand+1);
			theCommand[1] = '';
			(*theCommand)++;
			SetItem(theMenu, item, theCommand);
		}
	}
	#endif
	
	{
		MenuHandle mHandle;
		mHandle = GetMenu(kScriptMenu);
		/* Put the available scripts into the menu and check the currently selected menu */
		{
			scriptInfoPtr theScript;
			int theItem;
			
			OpenPlugIns();
			FindScripts();   /* Look for available login scripts */
			
			gScriptNumber=-1;
			HLock((Handle)gOurScripts);
			for (theScript=*gOurScripts,theItem=3; theScript->ourLength>0; 
					theItem++,theScript = (scriptInfoPtr)(theScript->ourLength+(long)theScript)) {
				Str255 theName;
				theName[0] = strlen(&theScript->fullName[theScript->printNameOffset]);
				BlockMove(&theScript->fullName[theScript->printNameOffset],&theName[1],theName[0]);
				AppendMenu(mHandle, theName);
				if (EqualString((StringPtr)(theScript->fullName), gScriptName, false, true)) {
					CheckItem(mHandle, 1, false);
					CheckItem(mHandle, theItem, true);
					gScriptNumber = theItem-2;
				}
			}
			HUnlock((Handle)gOurScripts);
		}
		InsertMenu(mHandle,-1);
	}
	
	DrawMenuBar();
}



/*----------------------------------------------------------------------------
	Init 
		
	Initializes the program. 	
----------------------------------------------------------------------------*/

void Init (void)
{
	EventRecord ev;
	CStr255 msg;
	Boolean mightHaveNewGroups = true;
	short sizeofTPrefRec;
	FSSpec **docList = nil;
	short numDocs;
	long systemVersion;
	long quickdrawVersion;
	OSErr err;
	
	extern Boolean TrapAvailable (unsigned long theTrap);	/* defined in dnr.c! */
	
	gLifeBoat = NewHandle(kLifeBoatSize);
	if (MemError() == noErr)
		gSinking = false;
	MoveHHi(gLifeBoat);
	HLock(gLifeBoat);
	
	InitGraf(&qd.thePort);
	InitFonts();
	InitWindows();
	InitMenus();
	TEInit();
	InitDialogs(nil);
	InitCursor();
	InitCursorCtl(nil);

	FlushEvents(everyEvent,0);
	EventAvail(everyEvent,&ev);

	if (!TrapAvailable(_Gestalt)) {
		ErrorMessage("NewsWatcher requires System 7.0 or later.");
		ExitToShell();
	}
	err = Gestalt(gestaltSystemVersion, &systemVersion);
	if (err != noErr || systemVersion < 0x0700) {
		ErrorMessage("NewsWatcher requires System 7.0 or later.");
		ExitToShell();
	}
	
	err = Gestalt(gestaltQuickdrawVersion, &quickdrawVersion);
	gHasColorQD = err == noErr && quickdrawVersion >= gestalt8BitQD;

	gDesktopExtent = (**GetGrayRgn()).rgnBBox;
	gDragRect = gDesktopExtent;
	InsetRect(&gDragRect,4,4);
	SetRect(&gWindLimits,kMinWindWidth,kMinWindHeight,6000,6000);
	gIBeamCurs = **(GetCursor(iBeamCursor));
	SpinCursor(0);
	
	sizeofTPrefRec = sizeof(TPrefRec);
	if (sizeofTPrefRec != kTPrefRecSize) {
		sprintf(msg, "sizeof(TPrefRec) is %d, should be %d", sizeofTPrefRec,
			kTPrefRecSize);
		ErrorMessage(msg);
		ExitToShell();
	}


/*
ErrorMessage("NewsWatcher 2.0d14 test version 4. Special test version! Please do not redistribute!");
*/
	
	InitializeAppleEvents();
	GetInitialDocList(&docList, &numDocs);
	LocatePrefsFile(docList, numDocs);
	FSMakeFSSpec(gPrefsFile.vRefNum, gPrefsFile.parID, kFilterName, &gFilterFile);
	gStartupOK = ReadPrefs();
	SetUpMenus();
	if (gPrefs.logActionsToFile) OpenLogFile();

	#ifdef FILTERS
		if (gStartupOK) gStartupOK = ReadKillFile();
	#endif

	if (gStartupOK) gStartupOK = StartNNTP();
	if (gStartupOK && gNumGroups == 0) {
		gStartupOK = ReadGroupsFromServer();
		mightHaveNewGroups = false;
	}
	
	if (gStartupOK) gStartupOK = CreateFullGroupListWindow();
	
	if (gStartupOK && mightHaveNewGroups && gPrefs.checkForNewGroups)
		CheckForNewGroups();

	if (gStartupOK) {
		if (gPrefs.autoFetchnewsrc) {
			DoGetGroupListFromHost(true);
			gCancel = false;
		}
		OpenDocList(docList, numDocs);
	}
	
	if (gStartupOK) CreateNewGroupListWindow();
	
	if (!gStartupOK) {
		EndNNTP();
		strcpy(msg, "An error occurred during the startup process, ");
		strcat(msg, "or you canceled the startup process. ");
		strcat(msg, "You may adjust NewsWatchers preferences, ");
		strcat(msg, "but you will not be able to read or post news.");
		ErrorMessage(msg);
	}
	
	gStartingUp = false;
}
